import pytest

from sirbot import SirBot
from sirbot.plugins.apscheduler import APSchedulerPlugin


@pytest.fixture
@pytest.mark.asyncio
async def bot():
    b = SirBot()
    b.load_plugin(APSchedulerPlugin())
    return b


class TestPluginAPscheduler:
    async def test_start(self, bot, test_server):
        await test_server(bot)
        assert isinstance(bot['plugins']['scheduler'], APSchedulerPlugin)



import os
import ujson
import logging
import asyncpg
import aiofiles

from aiocontext import async_contextmanager

LOG = logging.getLogger(__name__)


class PgPlugin:
    __name__ = 'pg'

    def __init__(self, *, sql_migration_directory=None, version=None, **kwargs):
        self.pool_kwargs = kwargs
        self.pool = None
        self.version = version

        if sql_migration_directory and not os.path.isabs(sql_migration_directory):
            self.sql_migration_directory = os.path.abspath(sql_migration_directory)
        else:
            self.sql_migration_directory = sql_migration_directory

    def load(self, sirbot):
        LOG.info('Loading postgres plugin')
        sirbot.on_startup.insert(0, self.startup)
        sirbot.on_shutdown.append(self.shutdown)

    async def startup(self, sirbot):
        self.pool = await asyncpg.create_pool(**self.pool_kwargs, init=self._init_connection)
        if self.sql_migration_directory and self.version:
            await self.migrate()

    async def shutdown(self, sirbot):
        await self.pool.close()

    @async_contextmanager
    async def connection(self):
        async with self.pool.acquire() as pg_con:
            yield pg_con

    async def migrate(self):
        LOG.info('Start of database migration')
        current_version = self.version.split('.')
        async with self.connection() as connection:
            old_version = await self._check_database_version(connection)

            async with connection.transaction():
                if old_version is None:
                    await self._init_database(connection)
                    old_version = (0, 0, 0)

                if old_version != current_version:
                    for version in self._find_update_version(start=old_version, end=current_version):
                        await self._execute_sql_file(connection, version)
                await self._update_db_version(connection, current_version)
                LOG.info('End of database migration')

    def _find_update_version(self, start, end):
        files = []
        if os.path.exists(self.sql_migration_directory):
            for file in os.listdir(self.sql_migration_directory):
                name, _ = os.path.splitext(file)
                file_version = name.split('.')
                if end >= file_version > start:
                    files.append(file_version)
            files = sorted(files)
        LOG.debug('Database migration versions: %s', files)
        return files

    async def _init_database(self, connection):
        LOG.debug('Executing initial migration')
        await connection.execute('''
                          CREATE TABLE metadata (db_version TEXT);
                          INSERT INTO metadata (db_version) VALUES ('0.0.0');
        ''')
        if os.path.exists(os.path.join(self.sql_migration_directory, 'init.sql')):
            await self._execute_sql_file(connection, 'init')

    async def _execute_sql_file(self, connection, version):
        LOG.debug('Database migration to version: %s', version)
        if version == 'init':
            file = 'init.sql'
        else:
            file = '{}.sql'.format('.'.join(version))

        async with aiofiles.open(os.path.join(self.sql_migration_directory, file), mode='r') as f:
            await connection.execute((await f.read()))
        LOG.debug('Database migration to version: %s OK', version)

    @staticmethod
    async def _check_database_version(connection):
        try:
            metadata = await connection.fetchrow('''SELECT * FROM metadata''')
            return metadata['db_version'].split('.')
        except asyncpg.exceptions.UndefinedTableError:
            LOG.debug('No "metadata" table found in database')

    @staticmethod
    async def _update_db_version(connection, version):
        await connection.execute('''UPDATE metadata SET db_version=$1''', '.'.join(version))

    async def _init_connection(self, connection):
        await connection.set_type_codec(
            'jsonb', encoder=self._json_encoder, decoder=self._json_decoder, schema='pg_catalog'
        )

    @staticmethod
    def _json_encoder(value):
        return ujson.dumps(value)

    @staticmethod
    def _json_decoder(value):
        return ujson.loads(value)



import re
import slack
import pytest
import asyncio
import asynctest

from unittest import mock

from sirbot import SirBot
from sirbot.plugins.slack import SlackPlugin


@pytest.fixture
@pytest.mark.asyncio
async def bot():
    b = SirBot()
    b.load_plugin(SlackPlugin(token='foo', verify='supersecuretoken', bot_user_id='baz', bot_id='boo',
                              admins=['aaa', 'bbb']))
    return b


@pytest.fixture
def find_bot_id_query():
    async def query(*args, **kwargs):
        return {
            "ok": True,
            "user": {
                "id": "W012A3CDE",
                "team_id": "T012AB3C4",
                "name": "spengler",
                "deleted": False,
                "color": "9f69e7",
                "real_name": "episod",
                "tz": "America\/Los_Angeles",
                "tz_label": "Pacific Daylight Time",
                "tz_offset": -25200,
                "profile": {
                    "avatar_hash": "ge3b51ca72de",
                    "status_text": "Print is dead",
                    "status_emoji": ":books:",
                    "real_name": "Egon Spengler",
                    "display_name": "spengler",
                    "real_name_normalized": "Egon Spengler",
                    "display_name_normalized": "spengler",
                    "email": "spengler@ghostbusters.example.com",
                    "image_24": "https:\/\/...\/avatar\/e3b51ca72dee4ef87916ae2b9240df50.jpg",
                    "image_32": "https:\/\/...\/avatar\/e3b51ca72dee4ef87916ae2b9240df50.jpg",
                    "image_48": "https:\/\/...\/avatar\/e3b51ca72dee4ef87916ae2b9240df50.jpg",
                    "image_72": "https:\/\/...\/avatar\/e3b51ca72dee4ef87916ae2b9240df50.jpg",
                    "image_192": "https:\/\/...\/avatar\/e3b51ca72dee4ef87916ae2b9240df50.jpg",
                    "image_512": "https:\/\/...\/avatar\/e3b51ca72dee4ef87916ae2b9240df50.jpg",
                    "team": "T012AB3C4",
                    "bot_id": "B00000000"
                },
                "is_admin": True,
                "is_owner": False,
                "is_primary_owner": False,
                "is_restricted": False,
                "is_ultra_restricted": False,
                "is_bot": True,
                "updated": 1502138686,
                "is_app_user": False,
                "has_2fa": False
            }
        }
    return query


class TestPluginSlack:
    async def test_start(self, bot, test_server):
        await test_server(bot)
        assert isinstance(bot['plugins']['slack'], SlackPlugin)

    async def test_start_no_bot_user_id(self, caplog):
        SlackPlugin(token='foo', verify='bar', bot_id='boo', admins=['aaa', 'bbb'])
        assert '`SLACK_BOT_USER_ID` not set' in caplog.text

    async def test_register_event(self, bot):
        async def handler():
            pass

        def handler2():
            pass

        bot['plugins']['slack'].on_event('team_join', handler)
        bot['plugins']['slack'].on_event('team_join', handler2)

        assert bot['plugins']['slack'].routers['event']._routes['team_join']['*']['*'][0][0] is handler
        assert asyncio.iscoroutinefunction(
            bot['plugins']['slack'].routers['event']._routes['team_join']['*']['*'][0][0])
        assert asyncio.iscoroutinefunction(
            bot['plugins']['slack'].routers['event']._routes['team_join']['*']['*'][1][0])

    async def test_register_command(self, bot):
        async def handler():
            pass

        def handler2():
            pass

        bot['plugins']['slack'].on_command('/hello', handler)
        bot['plugins']['slack'].on_command('/hello', handler2)

        assert bot['plugins']['slack'].routers['command']._routes['/hello'][0][0] is handler
        assert asyncio.iscoroutinefunction(bot['plugins']['slack'].routers['command']._routes['/hello'][0][0])
        assert asyncio.iscoroutinefunction(bot['plugins']['slack'].routers['command']._routes['/hello'][1][0])

    async def test_register_message(self, bot):
        async def handler():
            pass

        def handler2():
            pass

        msg = 'hello'
        bot['plugins']['slack'].on_message(msg, handler)
        bot['plugins']['slack'].on_message(msg, handler2)
        msg_compile = re.compile(msg)
        assert bot['plugins']['slack'].routers['message']._routes['*'][None][msg_compile][0][0] is handler
        assert asyncio.iscoroutinefunction(
            bot['plugins']['slack'].routers['message']._routes['*'][None][msg_compile][0][0])
        assert asyncio.iscoroutinefunction(
            bot['plugins']['slack'].routers['message']._routes['*'][None][msg_compile][1][0])

    async def test_register_admin_message_no_admin(self, caplog):
        async def handler():
            pass

        bot = SirBot()
        bot.load_plugin(SlackPlugin(token='foo', verify='bar', bot_user_id='baz', bot_id='boo'))
        bot['plugins']['slack'].on_message('hello', handler, admin=True)
        assert 'Slack admins ids are not set. Admin limited endpoint will not work.' in caplog.text

    async def test_register_action(self, bot):
        async def handler():
            pass

        def handler2():
            pass

        bot['plugins']['slack'].on_action('hello', handler)
        bot['plugins']['slack'].on_action('hello', handler2)

        assert bot['plugins']['slack'].routers['action']._routes['hello']['*'][0][0] is handler
        assert asyncio.iscoroutinefunction(bot['plugins']['slack'].routers['action']._routes['hello']['*'][0][0])
        assert asyncio.iscoroutinefunction(bot['plugins']['slack'].routers['action']._routes['hello']['*'][1][0])

    async def test_find_bot_id(self, bot, test_server, find_bot_id_query):
        await test_server(bot)
        bot['plugins']['slack'].api.query = find_bot_id_query
        await bot['plugins']['slack'].find_bot_id(bot)
        assert bot['plugins']['slack'].bot_id == 'B00000000'

    async def test_start_find_bot_id(self, test_server, find_bot_id_query):
        bot = SirBot()
        bot.load_plugin(SlackPlugin(token='foo', verify='bar', bot_user_id='baz'))
        bot['plugins']['slack'].api.query = find_bot_id_query
        await test_server(bot)
        assert bot['plugins']['slack'].bot_id == 'B00000000'


class TestPluginSlackEndpoints:
    async def test_incoming_event(self, bot, test_client, slack_event):
        client = await test_client(bot)
        r = await client.post('/slack/events', json=slack_event)
        assert r.status == 200

    async def test_incoming_command(self, bot, test_client, slack_command):
        client = await test_client(bot)
        r = await client.post('/slack/commands', data=slack_command)
        assert r.status == 200

    async def test_incoming_action(self, bot, test_client, slack_action):
        client = await test_client(bot)
        r = await client.post('/slack/actions', data=slack_action)
        assert r.status == 200

    async def test_incoming_event_wrong_token(self, bot, test_client, slack_event):
        bot['plugins']['slack'].verify = 'bar'
        client = await test_client(bot)
        r = await client.post('/slack/events', json=slack_event)
        assert r.status == 401

    async def test_incoming_command_wrong_token(self, bot, test_client, slack_command):
        bot['plugins']['slack'].verify = 'bar'
        client = await test_client(bot)
        r = await client.post('/slack/commands', data=slack_command)
        assert r.status == 401

    async def test_incoming_action_wrong_token(self, bot, test_client, slack_action):
        bot['plugins']['slack'].verify = 'bar'
        client = await test_client(bot)
        r = await client.post('/slack/actions', data=slack_action)
        assert r.status == 401

    async def test_incoming_event_error(self, bot, test_client, slack_event_only):
        async def handler(*args, **kwargs):
            raise RuntimeError()

        bot['plugins']['slack'].routers['event'].dispatch = mock.MagicMock(return_value=[(handler, {'wait': True}), ])
        bot['plugins']['slack'].routers['message'].dispatch = mock.MagicMock(return_value=[(handler, {'wait': True}), ])

        client = await test_client(bot)
        r = await client.post('/slack/events', json=slack_event_only)
        assert r.status == 500

    async def test_incoming_message_error(self, bot, test_client, slack_message):
        async def handler(*args, **kwargs):
            raise RuntimeError()

        bot['plugins']['slack'].routers['message'].dispatch = mock.MagicMock(
            return_value=[(handler, {'wait': True, 'mention': False, 'admin': False}), ])

        client = await test_client(bot)
        r = await client.post('/slack/events', json=slack_message)
        assert r.status == 500

    async def test_incoming_command_error(self, bot, test_client, slack_command):
        async def handler(*args, **kwargs):
            raise RuntimeError()

        bot['plugins']['slack'].routers['command'].dispatch = mock.MagicMock(return_value=[(handler, {'wait': True}), ])

        client = await test_client(bot)
        r = await client.post('/slack/commands', data=slack_command)
        assert r.status == 500

    async def test_incoming_action_error(self, bot, test_client, slack_action):
        async def handler(*args, **kwargs):
            raise RuntimeError()

        bot['plugins']['slack'].routers['action'].dispatch = mock.MagicMock(return_value=[(handler, {'wait': True}), ])

        client = await test_client(bot)
        r = await client.post('/slack/actions', data=slack_action)
        assert r.status == 500

    async def test_incoming_event_handler_arg(self, bot, test_client, slack_event):
        async def handler(event, app):
            assert app is bot
            assert isinstance(event, slack.events.Event)

        bot['plugins']['slack'].routers['event'].dispatch = mock.MagicMock(return_value=[(handler, {'wait': True}), ])

        client = await test_client(bot)
        r = await client.post('/slack/events', json=slack_event)
        assert r.status == 200

    async def test_incoming_message_handler_arg(self, bot, test_client, slack_message):
        async def handler(event, app):
            assert app is bot
            assert isinstance(event, slack.events.Message)

        bot['plugins']['slack'].routers['message'].dispatch = mock.MagicMock(
            return_value=[(handler, {'wait': True, 'mention': False, 'admin': False}), ])

        client = await test_client(bot)
        r = await client.post('/slack/events', json=slack_message)
        assert r.status == 200

    async def test_incoming_command_handler_arg(self, bot, test_client, slack_command):
        async def handler(command, app):
            assert app is bot
            assert isinstance(command, slack.commands.Command)

        bot['plugins']['slack'].routers['command'].dispatch = mock.MagicMock(return_value=[(handler, {'wait': True}), ])

        client = await test_client(bot)
        r = await client.post('/slack/commands', data=slack_command)
        assert r.status == 200

    async def test_incoming_action_handler_arg(self, bot, test_client, slack_action):
        async def handler(action, app):
            assert app is bot
            assert isinstance(action, slack.actions.Action)

        bot['plugins']['slack'].routers['action'].dispatch = mock.MagicMock(return_value=[(handler, {'wait': True}), ])

        client = await test_client(bot)
        r = await client.post('/slack/actions', data=slack_action)
        assert r.status == 200

    async def test_event_challenge(self, bot, test_client):

        client = await test_client(bot)
        r = await client.post('/slack/events', json={'token': 'supersecuretoken', 'challenge': 'abcdefghij'})
        data = await r.text()
        assert r.status == 200
        assert data == 'abcdefghij'

    async def test_event_challenge_wrong_token(self, bot, test_client):

        client = await test_client(bot)
        r = await client.post('/slack/events', json={'token': 'wrongsupersecuretoken', 'challenge': 'abcdefghij'})
        assert r.status == 500

    @pytest.mark.parametrize('slack_message', ('bot', ), indirect=True)
    async def test_message_from_bot(self, bot, test_client, slack_message):
        bot['plugins']['slack'].bot_id = 'B0AAA0A00'
        bot['plugins']['slack'].routers['message'].dispatch = mock.MagicMock()

        client = await test_client(bot)
        r = await client.post('/slack/events', json=slack_message)
        assert r.status == 200
        assert bot['plugins']['slack'].routers['message'].dispatch.call_count == 0

    @pytest.mark.parametrize('slack_message', ('bot', ), indirect=True)
    async def test_message_from_other_bot(self, bot, test_client, slack_message):
        bot['plugins']['slack'].bot_id = 'B0AAA0A01'
        bot['plugins']['slack'].routers['message'].dispatch = mock.MagicMock()

        client = await test_client(bot)
        r = await client.post('/slack/events', json=slack_message)
        assert r.status == 200
        assert bot['plugins']['slack'].routers['message'].dispatch.call_count == 1

    @pytest.mark.parametrize('slack_message', ('simple', ), indirect=True)
    async def test_admin_message_ok(self, bot, test_client, slack_message):
        handler = asynctest.CoroutineMock()
        bot['plugins']['slack'].admins = ['U000AA000', ]
        bot['plugins']['slack'].on_message('hello', handler, admin=True)

        client = await test_client(bot)
        r = await client.post('/slack/events', json=slack_message)
        assert r.status == 200
        assert handler.call_count == 1

    @pytest.mark.parametrize('slack_message', ('simple', ), indirect=True)
    async def test_admin_message_skip(self, bot, test_client, slack_message):
        handler = asynctest.CoroutineMock()
        bot['plugins']['slack'].admins = ['U000AA001', ]
        bot['plugins']['slack'].on_message('hello', handler, admin=True)

        client = await test_client(bot)
        r = await client.post('/slack/events', json=slack_message)
        assert r.status == 200
        assert handler.call_count == 0

    @pytest.mark.parametrize('slack_message', ('mention', ), indirect=True)
    async def test_message_mention_ok(self, bot, test_client, slack_message):
        handler = asynctest.CoroutineMock()
        bot['plugins']['slack'].bot_user_id = 'U0AAA0A00'
        bot['plugins']['slack'].on_message('hello world', handler, mention=True)

        client = await test_client(bot)
        r = await client.post('/slack/events', json=slack_message)
        assert r.status == 200
        assert handler.call_count == 1

    @pytest.mark.parametrize('slack_message', ('mention', ), indirect=True)
    async def test_message_mention_skip(self, bot, test_client, slack_message):
        handler = asynctest.CoroutineMock()
        bot['plugins']['slack'].bot_user_id = 'U0AAA0A01'
        bot['plugins']['slack'].on_message('hello world', handler, mention=True)

        client = await test_client(bot)
        r = await client.post('/slack/events', json=slack_message)
        assert r.status == 200
        assert handler.call_count == 0



import pytest

from sirbot import SirBot


@pytest.mark.asyncio
class TestSirBot:

    async def test_bot(self, test_server):
        bot = SirBot()
        await test_server(bot)

    async def test_load_plugin(self, test_server):
        class MyPlugin:
            __name__ = 'myplugin'

            def __init__(self):
                pass

            def load(self, test_bot):
                pass

        bot = SirBot()
        bot.load_plugin(MyPlugin())
        assert 'myplugin' in bot.plugins
        assert isinstance(bot['plugins']['myplugin'], MyPlugin)
        await test_server(bot)

    async def test_load_plugin_no_name(self):
        class MyPlugin:

            def __init__(self):
                pass

            def load(self, test_bot):
                pass

        bot = SirBot()
        with pytest.raises(AttributeError):
            bot.load_plugin(MyPlugin())
        await bot.shutdown()


class TestEndpoints:

    async def test_list_plugin_empty(self, test_client):
        bot = SirBot()
        client = await test_client(bot)
        rep = await client.get('/sirbot/plugins')
        data = await rep.json()
        assert data == {'plugins': []}

    async def test_list_plugin(self, test_client):
        class MyPlugin:
            __name__ = 'myplugin'

            def __init__(self):
                pass

            def load(self, test_bot):
                pass

        bot = SirBot()
        bot.load_plugin(MyPlugin())
        client = await test_client(bot)
        rep = await client.get('/sirbot/plugins')
        data = await rep.json()
        assert data == {'plugins': ['myplugin']}



import asyncio
import logging
import aiohttp.web

from . import endpoints

LOG = logging.getLogger(__name__)


class SirBot(aiohttp.web.Application):
    def __init__(self, user_agent=None, **kwargs):
        super().__init__(**kwargs)

        self.router.add_route('GET', '/sirbot/plugins', endpoints.plugins)

        self['plugins'] = dict()
        self['http_session'] = aiohttp.ClientSession(loop=kwargs.get('loop') or asyncio.get_event_loop())
        self['user_agent'] = user_agent or 'sir-bot-a-lot'

        self.on_shutdown.append(self.stop)

    def start(self, **kwargs):
        LOG.info('Starting SirBot')
        aiohttp.web.run_app(self, **kwargs)

    def load_plugin(self, plugin, name=None):
        name = name or plugin.__name__
        self['plugins'][name] = plugin
        plugin.load(self)

    async def stop(self, sirbot):
        self['http_session'].close()

    @property
    def plugins(self):
        return self['plugins']

    @property
    def http_session(self):
        return self['http_session']

    @property
    def user_agent(self):
        return self['user_agent']



import logging

from apscheduler.schedulers.asyncio import AsyncIOScheduler

LOG = logging.getLogger(__name__)


class APSchedulerPlugin:
    __name__ = 'scheduler'

    def __init__(self, **kwargs):
        self.scheduler = AsyncIOScheduler(**kwargs)

    def load(self, sirbot):
        LOG.info('Loading apscheduler plugin')
        sirbot.on_startup.append(self.start)

    async def start(self, sirbot):
        self.scheduler.start()



import json
import pytest

from sirbot import SirBot
from sirbot.plugins.github import GithubPlugin


@pytest.fixture
@pytest.mark.asyncio
async def bot():
    b = SirBot()
    b.load_plugin(GithubPlugin(verify='supersecrettoken'))
    return b


@pytest.fixture(params=['pr_merged', 'issue_closed'])
async def event(request):
    with open(f'tests/data/github/events/{request.param}.json') as f:
        body = json.loads(f.read())

    with open(f'tests/data/github/events/{request.param}.headers.json') as f:
        headers = json.loads(f.read())

    return body, headers


class TestPluginGithub:
    async def test_start(self, bot, test_server):
        await test_server(bot)
        assert isinstance(bot['plugins']['github'], GithubPlugin)

    async def test_incoming_event(self, bot, test_client, event):
        client = await test_client(bot)
        r = await client.post('/github', json=event[0], headers=event[1])
        assert r.status == 200

    async def test_incoming_event_401(self, bot, test_client, event):
        bot['plugins']['github'].verify = 'wrongsupersecrettoken'
        client = await test_client(bot)
        r = await client.post('/github', json=event[0], headers=event[1])
        assert r.status == 401

    async def test_incoming_event_handler_error(self, bot, test_client, event):
        async def handler(event, app):
            raise RuntimeError()

        bot['plugins']['github'].router.add(handler, event[1]['X-GitHub-Event'], action=event[0]['action'])
        client = await test_client(bot)
        r = await client.post('/github', json=event[0], headers=event[1])
        assert r.status == 500



import sys
import pathlib
import setuptools


if sys.version_info < (3, 6):
    raise RuntimeError('SirBot requires Python 3.6+')

LONG_DESCRIPTION = pathlib.Path('README.rst').read_text('utf-8')

requires = {
    'install': ['aiohttp', 'aiofiles', 'asyncpg', 'asyncio-contextmanager', 'slack-sansio', 'gidgethub', 'ujson',
                'apscheduler'],
    'doc': ['sphinx', 'sphinxcontrib-asyncio', 'sphinxcontrib-napoleon'],
    'tests': ['tox', 'pytest-runner', 'flake8', 'pytest>=3.3.0', 'coverage', 'pytest-coverage', 'pytest-asyncio',
              'asynctest', 'pytest-aiohttp', 'slack-sansio[dev]'],
    'full': []
}

requires['tests'].extend(requires['doc'])
requires['full'].extend(requires['install'])
requires['full'].extend(requires['tests'])


def find_version():
    with open("sirbot/__version__.py") as f:
        version = f.readlines()[-1].split('=')[-1].strip().strip("'").strip('"')
        if not version:
            raise RuntimeError('No version found')

    return version


setuptools.setup(
    name='sirbot',
    long_description=LONG_DESCRIPTION,
    description='Rewrite of sir-bot-a-lot',
    keywords=[
        'sirbot',
        'chatbot',
        'bot',
        'slack',
    ],
    packages=setuptools.find_packages(),
    zip_safe=True,
    python_requires='~=3.6',
    install_requires=requires['install'],
    setup_requires=[],
    tests_require=requires['tests'],
    extras_require=requires,
    # See: http://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],
    author=(
        'Ovv <contact@ovv.wtf>',
    ),
    author_email='contact@ovv.wtf',
    license='MIT',
    url='https://github.com/pyslackers/sir-bot-a-lot2',
    version=find_version(),
)



import asyncio
import logging

from aiohttp.web import Response
from slack.events import Event
from slack.actions import Action
from slack.commands import Command
from slack.exceptions import FailedVerification

LOG = logging.getLogger(__name__)


async def incoming_event(request):
    slack = request.app.plugins['slack']
    payload = await request.json()
    LOG.log(5, 'Incoming event payload: %s', payload)

    if 'challenge' in payload:
        if payload['token'] == slack.verify:
            return Response(body=payload['challenge'])
        else:
            return Response(status=500)

    try:
        event = Event.from_http(payload, verification_token=slack.verify)
    except FailedVerification:
        return Response(status=401)

    if event['type'] == 'message':
        return await _incoming_message(event, request)
    else:
        return await _wait_for_handler(slack.routers['event'], event, request.app)


async def _incoming_message(event, request):
    slack = request.app.plugins['slack']

    if slack.bot_id and (event.get('bot_id') == slack.bot_id or event.get('message', {}).get('bot_id') == slack.bot_id):
        return Response(status=200)

    LOG.debug('Incoming message: %s', event)
    text = event.get('text')
    if slack.bot_user_id and text:
        mention = slack.bot_user_id in event['text'] or event['channel'].startswith('D')
    else:
        mention = False

    if mention and text:
        event['text'] = event['text'].strip(f'<@{slack.bot_user_id}>').strip()

    coro = []
    for handler, configuration in slack.routers['message'].dispatch(event):
        if configuration['mention'] and not mention:
            continue
        elif configuration['admin'] and event['user'] not in slack.admins:
            continue

        f = asyncio.ensure_future(handler(event, request.app))
        if configuration['wait']:
            coro.append(f)
        else:
            f.add_done_callback(_callback)

    return await _run_coroutines(coro)


async def incoming_command(request):
    slack = request.app.plugins['slack']
    payload = await request.post()

    try:
        command = Command(payload, verification_token=slack.verify)
    except FailedVerification:
        return Response(status=401)

    LOG.debug('Incoming command: %s', command)
    return await _wait_for_handler(slack.routers['command'], command, request.app)


async def incoming_action(request):
    slack = request.app.plugins['slack']
    payload = await request.post()
    LOG.log(5, 'Incoming action payload: %s', payload)

    try:
        action = Action.from_http(payload, verification_token=slack.verify)
    except FailedVerification:
        return Response(status=401)

    LOG.debug('Incoming action: %s', action)
    return await _wait_for_handler(slack.routers['action'], action, request.app)


def _callback(f):
    try:
        f.result()
    except Exception as e:
        LOG.exception(e)


async def _wait_for_handler(router, event, app):
    coro = list()
    for handler, configuration in router.dispatch(event):
        f = asyncio.ensure_future(handler(event, app))
        if configuration['wait']:
            coro.append(f)
        else:
            f.add_done_callback(_callback)

    return await _run_coroutines(coro)


async def _run_coroutines(coroutines=None):
    try:
        if coroutines:
            futures, _ = await asyncio.wait(coroutines, return_when=asyncio.ALL_COMPLETED)
            for f in futures:
                f.result()
    except Exception as e:
        LOG.exception(e)
        return Response(status=500)

    return Response(status=200)



import os
import logging
import asyncio

from slack import methods
from slack.io.aiohttp import SlackAPI
from slack.actions import Router as ActionRouter
from slack.commands import Router as CommandRouter
from slack.events import EventRouter, MessageRouter

from . import endpoints

LOG = logging.getLogger(__name__)


class SlackPlugin:
    __name__ = 'slack'

    def __init__(self, *, token=None, verify=None, bot_id=None, bot_user_id=None, admins=None):
        self.api = None
        self.token = token or os.environ['SLACK_TOKEN']
        self.admins = admins or os.environ.get('SLACK_ADMINS', [])
        self.verify = verify or os.environ['SLACK_VERIFY']
        self.bot_id = bot_id or os.environ.get('SLACK_BOT_ID')
        self.bot_user_id = bot_user_id or os.environ.get('SLACK_BOT_USER_ID')
        self.handlers_option = {}

        if not self.bot_user_id:
            LOG.warning('`SLACK_BOT_USER_ID` not set. It is required for `on mention` routing and discarding '
                        'message coming from Sir Bot-a-lot to avoid loops.')

        self.routers = {
            'event': EventRouter(),
            'command': CommandRouter(),
            'message': MessageRouter(),
            'action': ActionRouter(),
        }

    def load(self, sirbot):
        LOG.info('Loading slack plugin')
        self.api = SlackAPI(session=sirbot.http_session, token=self.token)

        sirbot.router.add_route('POST', '/slack/events', endpoints.incoming_event)
        sirbot.router.add_route('POST', '/slack/commands', endpoints.incoming_command)
        sirbot.router.add_route('POST', '/slack/actions', endpoints.incoming_action)

        if self.bot_user_id and not self.bot_id:
            sirbot.on_startup.append(self.find_bot_id)

    def on_event(self, event_type, handler, wait=True):
        if not asyncio.iscoroutinefunction(handler):
            handler = asyncio.coroutine(handler)
        configuration = {'wait': wait}
        self.routers['event'].register(event_type, (handler, configuration))

    def on_command(self, command, handler, wait=True):
        if not asyncio.iscoroutinefunction(handler):
            handler = asyncio.coroutine(handler)
        configuration = {'wait': wait}
        self.routers['command'].register(command, (handler, configuration))

    def on_message(self, pattern, handler, mention=False, admin=False, wait=True, **kwargs):
        if not asyncio.iscoroutinefunction(handler):
            handler = asyncio.coroutine(handler)

        if admin and not self.admins:
            LOG.warning('Slack admins ids are not set. Admin limited endpoint will not work.')

        configuration = {'mention': mention, 'admin': admin, 'wait': wait}
        self.routers['message'].register(pattern=pattern, handler=(handler, configuration), **kwargs)

    def on_action(self, action, handler, name='*', wait=True):
        if not asyncio.iscoroutinefunction(handler):
            handler = asyncio.coroutine(handler)
        configuration = {'wait': wait}
        self.routers['action'].register(action, (handler, configuration), name)

    async def find_bot_id(self, app):
        rep = await self.api.query(url=methods.USERS_INFO, data={'user': self.bot_user_id})
        self.bot_id = rep['user']['profile']['bot_id']
        LOG.warning('`SLACK_BOT_ID` not set. For a faster start time set it to: "%s"', self.bot_id)
